/*
** Module   : ComboBoxGuiImpl.java
** Abstract : GUI combo-box implementation.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150421 Initial version.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.Color;
import java.awt.Font;
import java.awt.BasicStroke;
import java.awt.Stroke;
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.util.*;

/**
 * Specific GUI implementation of COMBO-BOX widget.
 */
@SuppressWarnings("unchecked")
public class ComboBoxGuiImpl
extends ComboBox<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Output manager */
   private GuiOutputManager screen;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Selected item box color. */
   private Color selectionBack;

   /** Selected dotted line color. */
   private Color selectionDots;

   /** Light shadow box color. */
   private Color lightShadow;

   /** Dark shadow box color. */
   private Color darkShadow;
   
   /** Very dark shadow box color. */
   private Color veryDarkShadow;
   
   /** Shift in X to draw text inside entry. */
   private static final int TEXT_SHIFT = 4;
   
   /** Dotted line pattern for focused button text frame. */
   private Stroke dotStroke;
   
   /** Dotted line specific parameter. */
   private static final float DOT_STEP = 1.0f;
   
   /** Pressed button state flag. */
   private boolean pressed = false;
   
   /** Image to draw drop down button. */
   private ImageGuiImpl imgButt = null; 
   
   /** Image to draw drop down button in disabled state. */
   private ImageGuiImpl imgButtDisabled = null; 
   
   /** Name of the image to use for drop-down list activation button. */
   private static final String DROP_DOWN_BUTTON_NAME = "btn-down-arrow";
   
   /** Name of the image to use for drop-down list activation button disabled sate. */
   private static final String DROP_DOWN_BUTTON_NAME_INS = "btn-down-arrow-ins";
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ComboBoxGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ComboBoxConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WidgetConfig wcfg)
   {
      super.initialize(id, wcfg);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // output manager
      screen = screen();
      cc = screen.coordinates();
      // font resolver
      gf = new GuiFontResolver(Window.resolveWindow(this), gd, config(), false);
      // initialize the shadow colors
      lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      selectionBack  = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      // the color below is negative to selectionBack
      selectionDots  = new Color(255 - selectionBack.getRed(),
                                 255 - selectionBack.getGreen(),
                                 255 - selectionBack.getBlue());

      // initialize dot pattern 
      dotStroke = new BasicStroke(ButtonGuiImpl.DOT_LINE_WIDTH, BasicStroke.CAP_BUTT,
                                  BasicStroke.JOIN_BEVEL, 0.0f, // this must be just valid
                                  new float[] {DOT_STEP, DOT_STEP}, DOT_STEP);

      // create required images to draw drop-down button
      imgButt = initDropDownButton(DROP_DOWN_BUTTON_NAME);
      imgButtDisabled = initDropDownButton(DROP_DOWN_BUTTON_NAME_INS);
      // register tooltip action listener
//      gd.registerTooltipWorker(this, gf.font().font, ToolTip.isActive(config().tooltip));
      // handle pending images to preload
//      loadPendingImages();
      // adjust image coordinates and width
//      recalcImagesCoords();
   }

   /**
    * Draw the combo-box. Called by this ComboBox's parent container.
    */
   @Override
   public void draw()
   {
      // combo-box options need to be refreshed before every painting
      final ComboBoxConfig cbbcfg = config();
      gc = new GuiColorResolver(gd, cbbcfg, SysColor.COLOR_MENU, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();
      
      // resolve font
      final String text = getLabelText();
      final FontCache fntCache = gf.font();

      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));
//      NativeRectangle clip = bcfg.noFocus && imgDown != null
//         ? clipRectangle(p, new NativeDimension(d.width, d.height))
//         : clipRectangle(p, new NativeDimension(d.width + 1, d.height + 1));
      
      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);

            boolean focused = hasFocus() && isEnabled();
            
            drawFieldBorder(po, d);
            drawEntryField(po, d, fntCache.font, focused);
            drawButton(po, d, pressed);
//            drawDropDownList(po, d, true);
            // button state flags
/*            boolean highlighted = highlighted();
            
            // flat button flag
            boolean flat = bcfg.noFocus && bcfg.flatButton;
            
            // having prepacked images
            boolean builtInImages = false; 

            // draw button itself
            if (hasImages() && !isDefaultButton())
            {
               // check for built-in imaged
               builtInImages = hasPrepackagedImages(bcfg);
               
               // button border is included in methods below
               if (builtInImages)
               {
                  drawPrepackagedImages(po, d, highlighted, pressed, flat);
               }
               else
               {
                  drawButtonImages(po, d, highlighted, pressed, flat);
               }
            }
            else
            {
               // draw button border
               drawButtonBorder(po, d, highlighted, pressed, flat);
               // abd the body
               drawButtonInternal(po, d, text, fntCache.font, highlighted, pressed, flat);
            }
            
            // focused button needs special highlight
            if (highlighted)
            {
               drawHighlight(po, d, gc.fgColor, builtInImages);
            }*/
         }
         
      });
   }

   
   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
/*      if (!hasFocus() || !isEnabled())
      {
         return;
      }

      Point origin = screenLocation();
      screen().at(origin);
      requestSync();*/
   }

   /**
    * Obtain the input focus.
    */
   @Override
   public void requestFocus()
   {
      if (isEnabled())
      {
         // generate the FOCUS_GAINED event
         super.requestFocus();

//         screen().at(screenLocation());
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  Redraw combo-box in pressed
    * state and request focus.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // push button effect
      if (isEnabled() && !pressed)
      {
         requestFocus();
         pressed = true;
         repaint();
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  Redraw combo-box in
    * released state.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // release button effect
      if (isEnabled() && pressed)
      {
         pressed = false;
         repaint();
      }
   }

   /**
    * Sets up initial image object for drop down button.
    * 
    * @param   def
    *          Image definition object to override the image under fixup.
    *
    * @return   Image object for the given name.
    */
   private ImageGuiImpl initDropDownButton(String name)
   {
      ImageGuiImpl imgResult = null;
      // get the image data for requested image
      byte[] imageData = ImageHelper.getPrepackagedImageData(name);
      // initialize the given image
      if (imageData != null)
      {
         imgResult = new ImageGuiImpl(WidgetId.nextID());
         imgResult.setImage(imageData);
         ((ImageConfig)imgResult.config()).transparent = true;
         // set up button location
         NativePoint location = imgResult.physicalLocation();
         // the size of the entry field
         NativeDimension dim = physicalDimension();
         // button is on the right corner of the entry field
         imgResult.setPhysicalLocation(location.x + dim.width - ButtonGuiImpl.PREBUILT_SIZE - 1,
                                       location.y + (dim.height - ButtonGuiImpl.PREBUILT_SIZE)/2);
      }
      return imgResult;
   }
   
   /**
    * Draw 3D border around the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    */
   private void drawFieldBorder(NativePoint orig, NativeDimension dim)
   {
      int x1 = orig.x;
      int x2 = orig.x + dim.width;
      int y1 = orig.y;
      int y2 = orig.y + dim.height;
      // drawing the frame around entry field
      
      gd.setColor(darkShadow);
      gd.drawLine(x1, y1, x1, y2);
      gd.drawLine(x1 + 1, y1, x2, y1);

      gd.setColor(gc.bgColor);
      gd.drawLine(x1 + 1, y2, x2, y2);
      gd.drawLine(x2, y2 - 1, x2, y1 + 1);

      gd.setColor(veryDarkShadow);
//      gd.drawRect(x1 + 1, y1 + 1, dim.width - 2, dim.height - 2);
      gd.drawLine(x1 + 1, y1 + 1, x1 + 1, y2 - 1);
      gd.drawLine(x1 + 2, y1 + 1, x2 - 1, y1 + 1);
   }
   
   /**
    * Draw entry field of the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    * @param   fnt
    *          Font to use for combo-box entry field.
    * @param   focused
    *          Is the combo-box focused.
    */
   private void drawEntryField(NativePoint orig, NativeDimension dim, Font fnt, boolean focused)
   {
      // coordinates of the field area
      int w = dim.width - ButtonGuiImpl.PREBUILT_SIZE - 4;
      int h = dim.height - 3; 

      // draw internal background
      gd.setColor(lightShadow);
      gd.drawRect(orig.x + 2, orig.y + 2, w, h );
      if (focused)
      {
         gd.setColor(selectionBack);
      }
      gd.fillRect(orig.x + 3, orig.y + 3, w - 1, h - 1);
      
      // get the text to draw with field
      String display = null;
      
      display = model().selected();
      
      if (display == null)
         display = new String("");

      // limit the data portion of the field to the maximum allowed size
      // in chui char coordinates are always whole numbers
      if (display.length() > (int) limit)
         display = display.substring(0, (int) limit);
      else if (display.length() < (int) limit)
         display = StringHelper.leftAlignText(display, (int) limit);

      // draw entry field text
      gd.setGuiFont(fnt);
//      int xt = orig.x + (dim.width - gd.getTextWidth(display, fnt))/2;
      // left-center alignment
//      int xt = orig.x + gd.getTextWidth(" ", fnt);
      int xt = orig.x + TEXT_SHIFT;
      int yt = orig.y + 1 + (dim.height + fnt.getSize())/2;
      gd.setFontStyle(fnt.getStyle());
      // enabled state controls text drawing
      if (isEnabled())
      {
         if (focused)
         {
            gd.setColor(lightShadow);
         }
         else
         {
            gd.setColor(gc.fgColor);
         }
         gd.drawString(display, xt, yt);
      }
      else
      {
         gd.setColor(lightShadow);
         gd.drawString(display, xt + 1, yt + 1);
         gd.setColor(darkShadow);
         gd.drawString(display, xt, yt);
      }

      // draw dotted rectangle around focused button text
      if (focused)
      {
         // set dotted line style
         gd.setLineStroke(dotStroke);
         gd.setColor(selectionDots);
         gd.drawRect(orig.x + 3, orig.y + 3, w - 2, h - 2);
         // restore default line style
         gd.setLineStroke(new BasicStroke());
      }
//      if (hasFocus() && isEnabled())
//         drawCaret();
   }

   /**
    * Draw drop down button of the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    * @param   pressed
    *          Flag indicating the button state, pressed or released.
    */
   private void drawButton(NativePoint orig, NativeDimension dim, boolean pressed)
   {
      // simple check to prevent NPE
      if (imgButt == null || imgButtDisabled == null)
      {
         return;
      }
      
      // coordinates of the button area
      int x1 = orig.x + dim.width - ButtonGuiImpl.PREBUILT_SIZE - 1;
      int x2 = orig.x + dim.width;
      int y1 = orig.y + 2;
      int y2 = orig.y + dim.height;

      // draw the light frame around image
      gd.setColor(gc.bgColor);
      gd.fillRect(x1, y1, x2 - x1, y2 - y1);
      
      // disabled state has separate inage to draw  
      if (pressed)
      {
         // for pressed state need 1 pixel right-down shift 
         NativePoint location = imgButt.physicalLocation();
         imgButt.setPhysicalLocation(location.x + 1, location.y + 1);
         imgButt.draw();
         // draw small button frame
         gd.setColor(darkShadow);
         gd.drawRect(location.x, y1, x2 - x1 - 1, y2 - y1 - 1);
         // and restore origin after draw
         imgButt.setPhysicalLocation(location.x, location.y);
      }
      else
      {
         // determine what image to draw
         ImageGuiImpl imgReleased = null;
         if (isEnabled())
         {
            imgReleased = imgButt;
         }
         else
         {
            imgReleased = imgButtDisabled; 
         }
         imgReleased.draw();
         // draw small button frame
         NativePoint location = imgReleased.physicalLocation();
         gd.setColor(lightShadow);
         gd.drawLine(location.x + 1, y1 + 2,
                     location.x + 1, y2 - 3);
         gd.drawLine(location.x + 1, y1 + 1,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE - 2, y1 + 1);

         gd.setColor(darkShadow);
         gd.drawLine(location.x + 1, y2 - 2,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE - 1, y2 - 2);
         gd.drawLine(location.x + ButtonGuiImpl.PREBUILT_SIZE - 1, y1 + 1,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE - 1, y2 - 3);

         gd.setColor(veryDarkShadow);
         gd.drawLine(location.x, y2 - 1,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE, y2 - 1);
         gd.drawLine(location.x + ButtonGuiImpl.PREBUILT_SIZE, y1,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE, y2 - 2);
      }
   }

   /**
    * Draw drop down button of the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    */
   private void drawDropDownList(NativePoint orig, NativeDimension dim)
   {
   }
}
